# encoding: utf-8
require "logstash/inputs/base"
require "logstash/namespace"
require "socket" # for Socket.gethostname
require "stud/interval"
require "rufus/scheduler"

# Periodically run a shell command and capture the whole output as an event.
#
# Notes:
#
# * The `script` field of this event will be the python script run.
# * The `api_key` field of this event will be the key used to access NGC.
#
class LogStash::Inputs::Ngc < LogStash::Inputs::Base

  config_name "ngc"

  default :codec, "json"

  # path to python script to run.
  config :script, :validate => :string, :required => true

  # ngc api key
  config :api_key, :validate => :string, :required => true

  # ngc auth url
  config :auth_url, :validate => :string, :required => true

  # ngc api url
  config :api_url, :validate => :string, :required => true

  # ngc env (stg | prod)
  config :env, :validate => :string, :required => true

  # jwt token
  config :jwt_token, :validate => :string

  # Interval to run the command. Value is in seconds.
  # Either `interval` or `schedule` option must be defined.
  config :interval, :validate => :number

  # Schedule of when to periodically run command, in Cron format
  # For example: "* * * * *" (execute command every minute, on the minute)
  # Either `interval` or `schedule` option must be defined.
  config :schedule, :validate => :string

  def register
    @logger.info("Registering Exec Input", :type => @type, :script => @script, :interval => @interval, :schedule => @schedule)
    @hostname = Socket.gethostname
    @io       = nil
    
    if (@interval.nil? && @schedule.nil?) || (@interval && @schedule)
      raise LogStash::ConfigurationError, "jdbc input: either 'interval' or 'schedule' option must be defined."
    end
  end # def register

  def run(queue)
    if @schedule
      @scheduler = Rufus::Scheduler.new(:max_work_threads => 1)
      @scheduler.cron @schedule do
        execute(queue)
      end
      @scheduler.join
    else
      while !stop?
        duration = execute(queue)
        wait_until_end_of_interval(duration)
      end # loop
    end
  end # def run

  def stop
    close_io()
    @scheduler.shutdown(:wait) if @scheduler
  end

  # Execute a given command
  # @param [String] A command string
  # @param [Array or Queue] A queue to append events to
  def execute(queue)
    start = Time.now
    output = exit_status = nil
    begin
      output, exit_status = run_script()
    rescue StandardError => e
      @logger.error("Error while running script",
        :script => @script, :e => e, :backtrace => e.backtrace)
    rescue Exception => e
      @logger.error("Exception while running script",
        :script => @script, :e => e, :backtrace => e.backtrace)
    end
    duration = Time.now - start
    @logger.debug? && @logger.debug("Command completed", :script => @script, :duration => duration)
    if output
      @codec.decode(output) do |decoded|

        @jwt_token = decoded.get("token")

        decoded.get("values").each do |val|
          event = LogStash::Event.new(val)
          decorate(event)
          event.set("[@metadata][host]", @hostname)
          event.set("[@metadata][duration]", duration)
          event.set("[@metadata][exit_status]", exit_status)
          queue << event
        end
      end
    end
    duration
  end

  private

  def run_script
    command = "python3 #{@script} -k #{@api_key} -e #{@env} --auth_url #{@auth_url} --api_url #{@api_url}"
    if @jwt_token
      command << " -t #{@jwt_token}"
    end

    @logger.debug? && @logger.debug("Running exec", :command => command)
      
    @io = IO.popen(command)
    output = @io.read
    @io.close # required in order to read $?
    exit_status = $?.exitstatus # should be threadsafe as per rb_thread_save_context
    [output, exit_status]
  ensure
    close_io()
  end

  # Close @io
  def close_io
    return if @io.nil? || @io.closed?
    @io.close
    @io = nil
  end

  # Wait until the end of the interval
  # @param [Integer] the duration of the last command executed
  def wait_until_end_of_interval(duration)
    # Sleep for the remainder of the interval, or 0 if the duration ran
    # longer than the interval.
    sleeptime = [0, @interval - duration].max
    if sleeptime > 0
      Stud.stoppable_sleep(sleeptime) { stop? }
    else
      @logger.warn("Execution ran longer than the interval. Skipping sleep.",
                   :script => @script, :duration => duration, :interval => @interval)
    end
  end


end # class LogStash::Inputs::Ngc
